use ::binding::{CDouble, CFloat, CInt};

use std::num::FpCategory;
use std::ops::Neg;

use ::constants::Constants;
use ::number::Number;
use ::one::One;
use ::zero::Zero;



// TODO: Double check these examples. Most are OK,
//       but I wanted to have f32 and f64 examples and
//       I'm not sure if all of them work correctly.
/// A trait that defines what is required to be considered
/// a Real number. [List of types of numbers][1]
///
/// This is meant to have all the similarities between
/// [f32][2] and [f64][3]. Most of the comments and documentation
/// were taken from the [rust-lang std docs][4].
///
/// [1]: https://en.wikipedia.org/wiki/List_of_types_of_numbers
/// [2]: https://doc.rust-lang.org/std/primitive.f32.html
/// [3]: https://doc.rust-lang.org/std/primitive.f64.html
/// [4]: https://doc.rust-lang.org/std/index.html
pub trait Real : Number + Constants + Neg<Output=Self>
{
   /// Smallest positive normal f64 value.
   const MIN_POSITIVE: Self;

   /// Difference between 1.0 and the next largest representable number.
   const EPSILON: Self;

   /// Infinity (∞).
   const INFINITY: Self;

   /// Negative infinity (-∞).
   const NEG_INFINITY: Self;

   /// Not a Number (NaN).
   const NAN: Self;



   /// Returns `true` if this value is `NaN` and false otherwise.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32 = f32::NAN;
   /// let nan64 = f64::NAN;
   /// let val32 = 7.0f32;
   /// let val64 = 7.0f64;
   ///
   /// assert!(nan32.is_nan());
   /// assert!(nan64.is_nan());
   /// assert!(!val32.is_nan());
   /// assert!(!val64.is_nan());
   /// ```
   fn is_nan(self) -> bool;

   /// Returns `true` if this number is neither infinite nor `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 7.0f32;
   /// let inf32: f32 = Real::INFINITY;
   /// let neg_inf32: f32 = Real::NEG_INFINITY;
   /// let nan32: f32 = Real::NAN;
   ///
   /// let val64 = 7.0f64;
   /// let inf64: f64 = Real::INFINITY;
   /// let neg_inf64: f64 = Real::NEG_INFINITY;
   /// let nan64: f64 = Real::NAN;
   ///
   /// assert!(val32.is_finite());
   /// assert!(!nan32.is_finite());
   /// assert!(!inf32.is_finite());
   /// assert!(!neg_inf32.is_finite());
   ///
   /// assert!(val64.is_finite());
   /// assert!(!nan64.is_finite());
   /// assert!(!inf64.is_finite());
   /// assert!(!neg_inf64.is_finite());
   /// ```
   fn is_finite(self) -> bool;

   /// Returns `true` if this value is positive
   /// infinity or negative infinity and false otherwise.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 7.0f32;
   /// let inf32: f32 = Real::INFINITY;
   /// let neg_inf32: f32 = Real::NEG_INFINITY;
   /// let nan32: f32 = Real::NAN;
   ///
   /// let val64 = 7.0f64;
   /// let inf64: f64 = Real::INFINITY;
   /// let neg_inf64: f64 = Real::NEG_INFINITY;
   /// let nan64: f64 = Real::NAN;
   ///
   /// assert!(!val32.is_infinite());
   /// assert!(!nan32.is_infinite());
   /// assert!(inf32.is_infinite());
   /// assert!(neg_inf32.is_infinite());
   ///
   /// assert!(!val64.is_infinite());
   /// assert!(!nan64.is_infinite());
   /// assert!(inf64.is_infinite());
   /// assert!(neg_inf64.is_infinite());
   /// ```
   fn is_infinite(self) -> bool;

   /// Returns `true` if the number is neither zero, infinite,
   /// subnormal, or `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let min32 = f32::MIN_POSITIVE;
   /// let max32 = f32::MAX;
   /// let lower_than_min32 = 1.0e-40_f32;
   /// let zero32 = 0.0f32;
   ///
   /// let min64 = f64::MIN_POSITIVE;
   /// let max64 = f64::MAX;
   /// let lower_than_min64 = 1.0e-40_f64;
   /// let zero64 = 0.0f64;
   ///
   /// // Values between `0` and `min` are Subnormal.
   /// assert!(min32.is_normal());
   /// assert!(max32.is_normal());
   /// assert!(!zero32.is_normal());
   /// assert!(!f32::NAN.is_normal());
   /// assert!(!f32::INFINITY.is_normal());
   /// assert!(!lower_than_min32.is_normal());
   ///
   /// assert!(min64.is_normal());
   /// assert!(max64.is_normal());
   /// assert!(!zero64.is_normal());
   /// assert!(!f64::NAN.is_normal());
   /// assert!(!f64::INFINITY.is_normal());
   /// //assert!(!lower_than_min64.is_normal());
   /// ```
   fn is_normal(self) -> bool;

   /// Returns `true` if `self` is negative, including `-0.0` and
   /// `Real::neg_infinity()`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32: f32 = f32::NAN;
   /// let f_val32 = 7.0f32;
   /// let g_val32 = -7.0f32;
   ///
   /// let nan64: f64 = f64::NAN;
   /// let f_val64 = 7.0f64;
   /// let g_val64 = -7.0f64;
   ///
   /// // Requires both tests to determine if is `NaN`
   /// assert!(!f_val32.is_sign_negative());
   /// assert!(g_val32.is_sign_negative());
   /// assert!(!nan32.is_sign_positive() && !nan32.is_sign_negative());
   ///
   /// assert!(!f_val64.is_sign_negative());
   /// assert!(g_val64.is_sign_negative());
   /// assert!(!nan64.is_sign_positive() && !nan64.is_sign_negative());
   /// ```
   fn is_sign_negative(self) -> bool;

   /// Returns `true` if `self` is positive, including `+0.0` and
   /// `Real::infinity()`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32: f32 = f32::NAN;
   /// let f_val32 = 7.0f32;
   /// let g_val32 = -7.0f32;
   ///
   /// let nan64: f64 = f64::NAN;
   /// let f_val64 = 7.0f64;
   /// let g_val64 = -7.0f64;
   ///
   /// // Requires both tests to determine if is `NaN`
   /// assert!(f_val32.is_sign_positive());
   /// assert!(!g_val32.is_sign_positive());
   /// assert!(!nan32.is_sign_positive() && !nan32.is_sign_negative());
   ///
   /// assert!(f_val64.is_sign_positive());
   /// assert!(!g_val64.is_sign_positive());
   /// assert!(!nan64.is_sign_positive() && !nan64.is_sign_negative());
   /// ```
   fn is_sign_positive(self) -> bool;

   /// Returns a number that represents the sign of `self`.
   ///
   /// - `1.0` if the number is positive, `+0.0` or `Real::infinity()`
   /// - `-1.0` if the number is negative, `-0.0` or `Real::neg_infinity()`
   /// - `Real::nan()` if the number is `Real::nan()`
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 3.5f32;
   ///
   /// let val64 = 3.5f64;
   ///
   /// assert_eq!(val32.signum(), 1.0f32);
   /// assert_eq!(f32::NEG_INFINITY.signum(), -1.0f32);
   /// assert!(f32::NAN.signum().is_nan());
   ///
   /// assert_eq!(val64.signum(), 1.0f64);
   /// assert_eq!(f64::NEG_INFINITY.signum(), -1.0f64);
   /// assert!(f64::NAN.signum().is_nan());
   /// ```
   fn signum(self) -> Self;

   /// Returns the floating point category of the number.
   /// If only one property is going to be tested, it is
   /// generally faster to use the specific predicate instead.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use std::num::FpCategory;
   /// use sigils::Real;
   ///
   /// let num32 = 12.4f32;
   /// let inf32 = f32::INFINITY;
   ///
   /// let num64 = 12.4f64;
   /// let inf64 = f64::INFINITY;
   ///
   /// assert_eq!(num32.classify(), FpCategory::Normal);
   /// assert_eq!(inf32.classify(), FpCategory::Infinite);
   ///
   /// assert_eq!(num64.classify(), FpCategory::Normal);
   /// assert_eq!(inf64.classify(), FpCategory::Infinite);
   /// ```
   fn classify(self) -> FpCategory;

   // TODO: Fix/check this example.
   /// Returns the mantissa, base 2 exponent, and sign as
   /// integers, respectively. The original number can be
   /// recovered by `sign * mantissa * 2 ^ exponent`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let num = 2.0f32;
   ///
   /// // (8388608, -22, 1)
   /// let (mantissa, exponent, sign) = Real::integer_decode(num);
   /// let sign_f = sign as f32;
   /// let mantissa_f = mantissa as f32;
   /// let exponent_f = num.powf(exponent as f32);
   ///
   /// // 1 * 8388608 * 2^(-22) == 2
   /// let abs_difference =
   ///    (sign_f * mantissa_f * exponent_f - num).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn integer_decode(self) -> (u64, i16, i8);

   /// Fused multiply-add. Computes `(self * a) + b` with
   /// only one rounding error. This produces a more accurate
   /// result with better performance than a separate
   /// multiplication operation followed by an add.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let m32 = 10.0f32;
   /// let x32 = 4.0f32;
   /// let b32 = 60.0f32;
   /// let abs_difference32 =
   ///    (m32.mul_add(x32, b32) - (m32*x32 + b32)).abs();
   ///
   /// let m64 = 10.0f64;
   /// let x64 = 4.0f64;
   /// let b64 = 60.0f64;
   /// let abs_difference64 =
   ///    (m64.mul_add(x64, b64) - (m64*x64 + b64)).abs();
   ///
   /// assert!(abs_difference32 < Real::EPSILON);
   /// assert!(abs_difference64 < Real::EPSILON);
   /// ```
   fn mul_add(self, a: Self, b: Self) -> Self;

   /// Raise a number to a floating point power.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 2.0f32;
   /// let abs_difference32 = (x32.powf(2.0f32) - x32*x32).abs();
   ///
   /// let x64 = 2.0f64;
   /// let abs_difference64 = (x64.powf(2.0f64) - x64*x64).abs();
   ///
   /// assert!(abs_difference32 < Real::EPSILON);
   /// assert!(abs_difference64 < Real::EPSILON);
   /// ```
   fn pow(self, n: Self) -> Self;

   /// Returns the smallest integer greater than or equal to a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.01f32;
   /// let g_val32 = 4.0f32;
   ///
   /// let f_val64 = 3.01f64;
   /// let g_val64 = 4.0f64;
   ///
   /// assert_eq!(f_val32.ceil(), 4.0f32);
   /// assert_eq!(g_val32.ceil(), 4.0f32);
   ///
   /// assert_eq!(f_val64.ceil(), 4.0f64);
   /// assert_eq!(g_val64.ceil(), 4.0f64);
   /// ```
   fn ceil(self) -> Self;

   /// Returns the largest integer less than or equal to a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.99f32;
   /// let g_val32 = 3.0f32;
   ///
   /// let f_val64 = 3.99f64;
   /// let g_val64 = 3.0f64;
   ///
   /// assert_eq!(f_val32.floor(), 3.0f32);
   /// assert_eq!(g_val32.floor(), 3.0f32);
   ///
   /// assert_eq!(f_val64.floor(), 3.0f64);
   /// assert_eq!(g_val64.floor(), 3.0f64);
   /// ```
   fn floor(self) -> Self;

   /// Returns the nearest integer to a number. Round
   /// half-way cases away from `0.0`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.3f32;
   /// let g_val32 = -3.3f32;
   ///
   /// let f_val64 = 3.3f64;
   /// let g_val64 = -3.3f64;
   ///
   /// assert_eq!(f_val32.round(), 3.0f32);
   /// assert_eq!(g_val32.round(), -3.0f32);
   ///
   /// assert_eq!(f_val64.round(), 3.0f64);
   /// assert_eq!(g_val64.round(), -3.0f64);
   /// ```
   fn round(self) -> Self;

   /// Return the integer part of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.3f32;
   /// let g_val32 = -3.7f32;
   ///
   /// let f_val64 = 3.3f64;
   /// let g_val64 = -3.7f64;
   ///
   /// assert_eq!(f_val32.trunc(), 3.0f32);
   /// assert_eq!(g_val32.trunc(), -3.0f32);
   ///
   /// assert_eq!(f_val64.trunc(), 3.0f64);
   /// assert_eq!(g_val64.trunc(), -3.0f64);
   /// ```
   fn trunc(self) -> Self;

   /// Returns the fractional part of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 3.5f32;
   /// let y32 = -3.5f32;
   /// let abs_difference_x32 = (x32.fract() - 0.5f32).abs();
   /// let abs_difference_y32 = (y32.fract() - (-0.5f32)).abs();
   ///
   /// let x64 = 3.5f64;
   /// let y64 = -3.5f64;
   /// let abs_difference_x64 = (x64.fract() - 0.5f64).abs();
   /// let abs_difference_y64 = (y64.fract() - (-0.5f64)).abs();
   ///
   /// assert!(abs_difference_x32 < Real::EPSILON);
   /// assert!(abs_difference_y32 < Real::EPSILON);
   ///
   /// assert!(abs_difference_x64 < Real::EPSILON);
   /// assert!(abs_difference_y64 < Real::EPSILON);
   /// ```
   fn fract(self) -> Self;

   /// Computes the absolute value of `self`. Returns `Real::nan()` if the
   /// number is `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let x32 = 3.5f32;
   /// let y32 = -3.5f32;
   /// let abs_difference_x32 = (x32.abs() - x32).abs();
   /// let abs_difference_y32 = (y32.abs() - (-y32)).abs();
   ///
   /// let x64 = 3.5f64;
   /// let y64 = -3.5f64;
   /// let abs_difference_x64 = (x64.abs() - x64).abs();
   /// let abs_difference_y64 = (y64.abs() - (-y64)).abs();
   ///
   /// assert!(abs_difference_x32 < Real::EPSILON);
   /// assert!(abs_difference_y32 < Real::EPSILON);
   /// assert!(f32::NAN.abs().is_nan());
   ///
   /// assert!(abs_difference_x64 < Real::EPSILON);
   /// assert!(abs_difference_y64 < Real::EPSILON);
   /// assert!(f64::NAN.abs().is_nan());
   /// ```
   fn abs(self) -> Self;

   /// Take the reciprocal (inverse) of a number, `1/x`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 2.0f32;
   /// let abs_difference32 =
   ///    (x32.recip() - (1.0f32/x32)).abs();
   ///
   /// let x64 = 2.0f64;
   /// let abs_difference64 =
   ///    (x64.recip() - (1.0f64/x64)).abs();
   ///
   /// assert!(abs_difference32 < Real::EPSILON);
   /// assert!(abs_difference64 < Real::EPSILON);
   /// ```
   fn recip(self) -> Self;

   /// Take the square root of a number.
   ///
   /// Returns NaN if `self` is a negative number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let positive = 4.0f32;
   /// let negative = -4.0f32;
   ///
   /// let abs_difference = (positive.sqrt() - 2.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// assert!(negative.sqrt().is_nan());
   /// ```
   fn sqrt(self) -> Self;

   /// Take the cubic root of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x: f64;
   /// let abs_difference: f64;
   ///
   /// // x^(1/3) - 2 == 0
   /// x = 8.0;
   /// abs_difference = (x.cbrt() - 2.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn cbrt(self) -> Self;

   /// Returns `e^(self)`, (the exponential function).
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let one = 1.0f64;
   /// // e^1
   /// let e = one.exp();
   ///
   /// // ln(e) - 1 == 0
   /// let abs_difference = (e.ln() - 1.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn exp(self) -> Self;

   /// Returns `2^(self)`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f = 2.0f64;
   ///
   /// // 2^2 - 4 == 0
   /// let abs_difference = (f.exp2() - 4.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn exp2(self) -> Self;

   /// Returns `e^(self) - 1` in a way that is accurate even if the
   /// number is close to zero.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x: f64;
   /// let abs_difference: f64;
   ///
   /// // e^(ln(7)) - 1
   /// x = 7.0;
   /// abs_difference = (x.ln().exp_m1() - 6.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn exp_m1(self) -> Self;

   /// Returns the natural logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let one = 1.0f64;
   /// // e^1
   /// let e = one.exp();
   ///
   /// // ln(e) - 1 == 0
   /// let abs_difference = (e.ln() - 1.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn ln(self) -> Self;

   /// Returns the logarithm of the number with respect to an arbitrary base.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let ten = 10.0f32;
   /// let two = 2.0f32;
   ///
   /// // log10(10) - 1 == 0
   /// let abs_difference_10 = (ten.log(10.0) - 1.0).abs();
   ///
   /// // log2(2) - 1 == 0
   /// let abs_difference_2 = (two.log(2.0) - 1.0).abs();
   ///
   /// assert!(abs_difference_10 < Real::EPSILON);
   /// assert!(abs_difference_2 < Real::EPSILON);
   /// ```
   fn log(self, base: Self) -> Self;

   /// Returns the base 2 logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let two = 2.0f32;
   ///
   /// // log2(2) - 1 == 0
   /// let abs_difference = (two.log2() - 1.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn log2(self) -> Self;

   /// Returns the base 10 logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let ten = 10.0f64;
   ///
   /// // log10(10) - 1 == 0
   /// let abs_difference = (ten.log10() - 1.0).abs();
   ///
   /// assert!(abs_difference < Real::EPSILON);
   /// ```
   fn log10(self) -> Self;

   /// Returns `ln(1+n)` (natural logarithm) more accurately than if
   /// the operations were performed separately.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// // ln(1 + (e - 1)) == ln(e) == 1
   /// let mut x32: f32 = Constants::E;
   /// x32 -= 1.0f32;
   /// let abs_difference32 = (x32.ln_1p() - 1.0f32).abs();
   ///
   /// let mut x64: f64 = Constants::E;
   /// x64 -= 1.0f64;
   /// let abs_difference64 = (x64.ln_1p() - 1.0f64).abs();
   ///
   /// //assert!(abs_difference32 < Real::EPSILON);
   /// assert!(abs_difference64 < Real::EPSILON);
   /// ```
   fn ln_1p(self) -> Self;
}



impl Real for f32
{
   ///
   const MIN_POSITIVE: Self = 1.17549435e-38f32;

   ///
   const EPSILON: Self = 1.19209290e-07f32;

   ///
   const INFINITY: Self = 1.0f32 / 0.0f32;

   ///
   const NEG_INFINITY: Self = -1.0f32 / 0.0f32;

   ///
   const NAN: Self = 0.0f32 / 0.0f32;



   fn is_nan(self) -> bool
   {
      self == Self::NAN
   }

   fn is_finite(self) -> bool
   {
      !self.is_infinite() && self != Self::NAN
   }

   fn is_infinite(self) -> bool
   {
      self == Self::INFINITY || self == Self::NEG_INFINITY
   }

   fn is_normal(self) -> bool
   {
      self.is_finite() && self != Self::zero() &&
      self <= -Self::MIN_POSITIVE && self >= Self::MIN_POSITIVE
   }

   fn is_sign_negative(self) -> bool
   {
      self.is_finite() && self < Self::zero()
   }

   fn is_sign_positive(self) -> bool
   {
      self.is_finite() && self > Self::zero()
   }

   fn signum(self) -> Self
   {
      if self.is_nan() == true
      {
         Self::NAN
      }
      else if self >= Self::zero() || self == Self::INFINITY
      {
         Self::one()
      }
      else if self <= -Self::zero() || self == Self::NEG_INFINITY
      {
         -Self::one()
      }
      else
      {
         Self::zero()
      }
   }

   fn classify(self) -> FpCategory
   {
      <Self>::classify(self)
   }

   fn integer_decode(self) -> (u64, i16, i8)
   {
      let bits: u32;
      let sign: i8;
      let mantissa: u32;
      let mut exponent: i16;

      bits = unsafe { ::std::mem::transmute(self) };
      if bits >> 31 == 0
      {
         sign = 1;
      }
      else
      {
         sign = -1;
      };

      exponent = ((bits >> 23) & 0xff) as i16;
      if exponent == 0
      {
         mantissa = (bits & 0x7fffff) << 1;
      }
      else
      {
         mantissa = (bits & 0x7fffff) | 0x800000;
      }

      // Exponent bias + mantissa shift
      exponent -= 127 + 23;

      (mantissa as u64, exponent, sign)
   }

   fn mul_add(self, a: Self, b: Self) -> Self
   {
      unsafe
      {
         ::pact::fmaf(self as CFloat, a as CFloat, b as CFloat) as Self
      }
   }

   fn pow(self, n: Self) -> Self
   {
      unsafe
      {
         ::pact::powf(self as CFloat, n as CFloat) as Self
      }
   }

   fn ceil(self) -> Self
   {
      unsafe
      {
         ::pact::ceilf(self as CFloat) as Self
      }
   }

   fn floor(self) -> Self
   {
      unsafe
      {
         ::pact::floorf(self as CFloat) as Self
      }
   }

   fn round(self) -> Self
   {
      unsafe
      {
         ::pact::roundf(self as CFloat) as Self
      }
   }

   fn trunc(self) -> Self
   {
      unsafe
      {
         ::pact::truncf(self as CFloat) as Self
      }
   }

   fn fract(self) -> Self
   {
      let mut integer: CInt;

      integer = 0;

      unsafe
      {
         ::pact::modff(self as CFloat, &mut integer) as Self
      }
   }

   fn abs(self) -> Self
   {
      unsafe
      {
         ::pact::fabsf(self as CFloat) as Self
      }
   }

   fn recip(self) -> Self
   {
      1.0f32 / self
   }

   fn sqrt(self) -> Self
   {
      unsafe
      {
         ::pact::sqrtf(self as CFloat) as Self
      }
   }

   fn cbrt(self) -> Self
   {
      unsafe
      {
         ::pact::cbrtf(self as CFloat) as Self
      }
   }

   fn exp(self) -> Self
   {
      unsafe
      {
         ::pact::expf(self as CFloat) as Self
      }
   }

   fn exp2(self) -> Self
   {
      unsafe
      {
         ::pact::exp2f(self as CFloat) as Self
      }
   }

   fn exp_m1(self) -> Self
   {
      unsafe
      {
         ::pact::expm1f(self as CFloat) as Self
      }
   }

   fn ln(self) -> Self
   {
      unsafe
      {
         ::pact::logf(self as CFloat) as Self
      }
   }

   fn log(self, base: Self) -> Self
   {
      self.ln() / base.ln()
   }

   fn log2(self) -> Self
   {
      unsafe
      {
         ::pact::log2f(self as CFloat) as Self
      }
   }

   fn log10(self) -> Self
   {
      unsafe
      {
         ::pact::log10f(self as CFloat) as Self
      }
   }

   fn ln_1p(self) -> Self
   {
      unsafe
      {
         ::pact::log1pf(self as CFloat) as Self
      }
   }
}

impl Real for f64
{
   ///
   const MIN_POSITIVE: Self = 2.2250738585072014e-308f64;

   ///
   const EPSILON: Self = 2.2204460492503131e-16f64;

   ///
   const INFINITY: Self = 1.0f64 / 0.0f64;

   ///
   const NEG_INFINITY: Self = -1.0f64 / 0.0f64;

   ///
   const NAN: Self = 0.0f64 / 0.0f64;



   fn is_nan(self) -> bool
   {
      self == Self::NAN
   }

   fn is_finite(self) -> bool
   {
      !self.is_infinite() && self != Self::NAN
   }

   fn is_infinite(self) -> bool
   {
      self == Self::INFINITY || self == Self::NEG_INFINITY
   }

   fn is_normal(self) -> bool
   {
      self.is_finite() && self != Self::zero() &&
      self <= -Self::MIN_POSITIVE && self >= Self::MIN_POSITIVE
   }

   fn is_sign_negative(self) -> bool
   {
      self.is_finite() && self < Self::zero()
   }

   fn is_sign_positive(self) -> bool
   {
      self.is_finite() && self > Self::zero()
   }

   fn signum(self) -> Self
   {
      if self.is_nan() == true
      {
         Self::NAN
      }
      else if self >= Self::zero() || self == Self::INFINITY
      {
         Self::one()
      }
      else if self <= -Self::zero() || self == Self::NEG_INFINITY
      {
         -Self::one()
      }
      else
      {
         Self::zero()
      }
   }

   fn classify(self) -> FpCategory
   {
      <Self>::classify(self)
   }

   fn integer_decode(self) -> (u64, i16, i8)
   {
      let bits: u64;
      let sign: i8;
      let mantissa: u64;
      let mut exponent: i16;

      bits = unsafe { ::std::mem::transmute(self) };
      if bits >> 63 == 0
      {
         sign = 1;
      }
      else
      {
         sign = -1;
      }

      exponent = ((bits >> 52) & 0x7ff) as i16;
      if exponent == 0
      {
        mantissa = (bits & 0xfffffffffffff) << 1;
      }
      else
      {
        mantissa = (bits & 0xfffffffffffff) | 0x10000000000000;
      }

      // Exponent bias + mantissa shift
      exponent -= 1023 + 52;

      (mantissa, exponent, sign)
   }

   fn mul_add(self, a: Self, b: Self) -> Self
   {
      unsafe
      {
         ::pact::fma(self as CDouble, a as CDouble, b as CDouble) as Self
      }
   }

   fn pow(self, n: Self) -> Self
   {
      unsafe
      {
         ::pact::pow(self as CDouble, n as CDouble) as Self
      }
   }

   fn ceil(self) -> Self
   {
      unsafe
      {
         ::pact::ceil(self as CDouble) as Self
      }
   }

   fn floor(self) -> Self
   {
      unsafe
      {
         ::pact::floor(self as CDouble) as Self
      }
   }

   fn round(self) -> Self
   {
      unsafe
      {
         ::pact::round(self as CDouble) as Self
      }
   }

   fn trunc(self) -> Self
   {
      unsafe
      {
         ::pact::trunc(self as CDouble) as Self
      }
   }

   fn fract(self) -> Self
   {
      let mut integer: CInt;

      integer = 0;

      unsafe
      {
         ::pact::modf(self as CDouble, &mut integer) as Self
      }
   }

   fn abs(self) -> Self
   {
      unsafe
      {
         ::pact::fabs(self as CDouble) as Self
      }
   }

   fn recip(self) -> Self
   {
      1.0f64 / self
   }

   fn sqrt(self) -> Self
   {
      unsafe
      {
         ::pact::sqrt(self as CDouble) as Self
      }
   }

   fn cbrt(self) -> Self
   {
      unsafe
      {
         ::pact::cbrt(self as CDouble) as Self
      }
   }

   fn exp(self) -> Self
   {
      unsafe
      {
         ::pact::exp(self as CDouble) as Self
      }
   }

   fn exp2(self) -> Self
   {
      unsafe
      {
         ::pact::exp2(self as CDouble) as Self
      }
   }

   fn exp_m1(self) -> Self
   {
      unsafe
      {
         ::pact::expm1(self as CDouble) as Self
      }
   }

   fn ln(self) -> Self
   {
      unsafe
      {
         ::pact::log(self as CDouble) as Self
      }
   }

   fn log(self, base: Self) -> Self
   {
      self.ln() / base.ln()
   }

   fn log2(self) -> Self
   {
      unsafe
      {
         ::pact::log2(self as CDouble) as Self
      }
   }

   fn log10(self) -> Self
   {
      unsafe
      {
         ::pact::log10(self as CDouble) as Self
      }
   }

   fn ln_1p(self) -> Self
   {
      unsafe
      {
         ::pact::log1p(self as CDouble) as Self
      }
   }
}
